<?php
namespace Privy\ORM;

/**
 * Privy.com Zen Cart Plugin
 * Copyright (C) 2015 Privy.com
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

abstract class Model {
    /**
     * Returns a new instance of Model with the given properties.
     *
     * @param array $properties - an associative array of properties
     * @return Model
     */
    static final function build($properties=NULL) {
        if ($properties == NULL) {
            $properties = static::defaultProperties();
        }

        $class = get_called_class();

        $obj = new $class();
        $obj->setProperties($properties);
        return $obj;
    }

    /**
     * Returns `columnName` prefixed by `COLUMN_PREFIX`.
     * @param string $columnName
     * @param string $sep - the separator used to produce the column name
     * @return string
     */
    final function c($columnName, $sep = "_") {
        return static::$COLUMN_PREFIX . $sep . $columnName;
    }

    /**
     * Implement this to define a default object for `::build`.
     * @return array - an associative array of properties
     */
    abstract static function defaultProperties();

    /**
     * Implement this to initialize the column names.
     * @return void
     */
    abstract static function init();

    /**
     * @return boolean - true if this instance of Model hasn't been saved.
     */
    final function isNewRecord() {
        return ($this->id == NULL);
    }

    /**
     * @return string - a SQL statement to insert this instance into `TABLE_NAME`.
     */
    protected final function insertStatement() {
        return "INSERT INTO " . static::$TABLE_NAME
                              . " (" . implode(", ", static::$COLUMN_NAMES) . ") "
                              . " values (" . implode(", ", $this->toArray()) . ");";
    }

    /**
     * Implement this. See `Model.findWhere`.
     */
    abstract static function find();

    /**
     * @param Model - a default instance to create
     * @return Model - if an instance was found or created,
     *  or NULL if one could not be created.
     */
    static final function findOrCreate($defaultObj=NULL) {
        $obj = forward_static_call_array(array(get_called_class(), "find"), func_get_args());
        if ($obj != NULL) {
            return $obj;
        } else {
            $obj = ($defaultObj != NULL) ? $defaultObj : static::build();
            if ($obj->save()) {
                return $obj;
            } else {
                return NULL;
            }
        }
    }

    /**
     * Find a Model satisfying the given conditions.
     *
     * @param conditions: string
     * @return \Privy\ORM\Model - if an instance was found, or
     *  NULL if the query produced no results.
     */
    static final function findWhere($conditions) {
        global $db;

        $query  = "SELECT * FROM " . static::$TABLE_NAME . " WHERE " . $conditions;

        $result = $db->Execute($query);
        if ($result->RecordCount() > 0) {
            return static::build($result->fields);
        } else {
            return NULL;
        }
    }

    /**
     * @param string $name
     * @param boolean $mapName
     * @return mixed
     */
    final function getProperty($name, $mapName=true) {
        if ($mapName) {
            $name = $this->mappedName($name);
        }

        return $this->$name;
    }

    /**
     * @param string $name - the simple column name
     * @return string - the fully qualified column name
     */
    final function mappedName($name) {
        $mappedName = static::c($name);
        if (in_array($mappedName, static::$COLUMN_NAMES)) {
            return $mappedName;
        } else if (in_array($name, static::$COLUMN_NAMES)) {
            return $name;
        } else {
            return NULL;
        }
    }

    /**
     * @return boolean - true if the save was successful, otherwise false.
     */
    final function save() {
        global $db;

        if ($this->isNewRecord()) {
            $stmt = $this->insertStatement();
        } else {
            $stmt = $this->updateStatement();
        }

        $res = $db->Execute($stmt);
        if ($this->isNewRecord()) {
            $this->setProperties(array("id" => $db->insert_ID()));
        }

        return ($res->resource === true);
    }

    /**
     * @param array $properties - an associative array of properties
     * @param bool $mapNames
     */
    final function setProperties($properties, $mapNames=true) {
        foreach ($properties as $key => $val) {
            if ($mapNames) {
                $name = $this->mappedName($key);
                if ($name == NULL) {
                    continue;
                }
            } else {
                $name = $key;
            }

            $this->$name = $val;
        }
    }

    /**
     * @param boolean $quote - if true, then the properties will be wrapped
     *  in single quotes.
     * @return array - a list of properties
     */
    final function toArray($quote=true) {
        $arr = array();
        foreach (static::$COLUMN_NAMES as $column) {
            $val = $this->$column;
            $arr[] = ($quote ? "'$val'" : "$val");
        }

        return $arr;
    }

    /**
     * @return string - a SQL statement to insert this instance into `TABLE_NAME`.
     */
    protected final function updateStatement() {
        return "UPDATE " . static::$TABLE_NAME
                         . " SET " . \Privy\Utils\join(static::$COLUMN_NAMES, $this->toArray(), " = ", ", ")
                         . " WHERE " . $this->c("id") . ";";
    }
}